"""Organize your data in a spatial grid system for CALS, HTML, Formex4, Office Open XML tables conversion"""

__version__ = "0.4.4"
__author__ = "Laurent LAPORTE <laurent.laporte.pro@gmail.com>"
__date__ = "2021-11-10"
__credit__ = "(c) 2018-2021 Laurent LAPORTE"
